# sage_setup: distribution = sagemath-categories
r"""
Filtered Modules With Basis

A *filtered module with basis* over a ring `R` means
(for the purpose of this code) a filtered `R`-module `M`
with filtration `(F_i)_{i \in I}` (typically `I = \NN`)
endowed with a basis `(b_j)_{j \in J}` of `M` and a partition
`J = \bigsqcup_{i \in I} J_i` of the set `J` (it is allowed
that some `J_i` are empty) such that for every `n \in I`,
the subfamily `(b_j)_{j \in U_n}`, where
`U_n = \bigcup_{i \leq n} J_i`, is a basis of the
`R`-submodule `F_n`.

For every `i \in I`, the `R`-submodule of `M` spanned by
`(b_j)_{j \in J_i}` is called the `i`-*th graded component*
(aka the `i`-*th homogeneous component*) of the filtered
module with basis `M`; the elements of this submodule are
referred to as *homogeneous elements of degree* `i`.

See the class documentation
:class:`~sage.categories.filtered_modules_with_basis.FilteredModulesWithBasis`
for further details.
"""
#*****************************************************************************
#  Copyright (C) 2014 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.categories.filtered_modules import FilteredModulesCategory
from sage.misc.abstract_method import abstract_method
from sage.misc.cachefunc import cached_method
from sage.categories.subobjects import SubobjectsCategory
from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring


class FilteredModulesWithBasis(FilteredModulesCategory):
    r"""
    The category of filtered modules with a distinguished basis.

    A *filtered module with basis* over a ring `R` means
    (for the purpose of this code) a filtered `R`-module `M`
    with filtration `(F_i)_{i \in I}` (typically `I = \NN`)
    endowed with a basis `(b_j)_{j \in J}` of `M` and a partition
    `J = \bigsqcup_{i \in I} J_i` of the set `J` (it is allowed
    that some `J_i` are empty) such that for every `n \in I`,
    the subfamily `(b_j)_{j \in U_n}`, where
    `U_n = \bigcup_{i \leq n} J_i`, is a basis of the
    `R`-submodule `F_n`.

    For every `i \in I`, the `R`-submodule of `M` spanned by
    `(b_j)_{j \in J_i}` is called the `i`-*th graded component*
    (aka the `i`-*th homogeneous component*) of the filtered
    module with basis `M`; the elements of this submodule are
    referred to as *homogeneous elements of degree* `i`.
    The `R`-module `M` is the direct sum of its `i`-th graded
    components over all `i \in I`, and thus becomes a graded
    `R`-module with basis.
    Conversely, any graded `R`-module with basis canonically
    becomes a filtered `R`-module with basis (by defining
    `F_n = \bigoplus_{i \leq n} G_i` where `G_i` is the `i`-th
    graded component, and defining `J_i` as the indexing set
    of the basis of the `i`-th graded component). Hence, the
    notion of a filtered `R`-module with basis is equivalent
    to the notion of a graded `R`-module with basis.

    However, the *category* of filtered `R`-modules with basis is not
    the category of graded `R`-modules with basis. Indeed, the *morphisms*
    of filtered `R`-modules with basis are defined to be morphisms of
    `R`-modules which send each `F_n` of the domain to the corresponding
    `F_n` of the target; in contrast, the morphisms of graded `R`-modules
    with basis must preserve each homogeneous component. Also,
    the notion of a filtered algebra with basis differs from
    that of a graded algebra with basis.

    .. NOTE::

        Currently, to make use of the functionality of this class,
        an instance of ``FilteredModulesWithBasis`` should fulfill
        the contract of a :class:`CombinatorialFreeModule` (most
        likely by inheriting from it). It should also have the
        indexing set `J` encoded as its ``_indices`` attribute,
        and ``_indices.subset(size=i)`` should yield the subset
        `J_i` (as an iterable). If the latter conditions are not
        satisfied, then :meth:`basis` must be overridden.

    .. NOTE::

        One should implement a ``degree_on_basis`` method in the parent
        class in order to fully utilize the methods of this category.
        This might become a required abstract method in the future.

    EXAMPLES::

        sage: C = ModulesWithBasis(ZZ).Filtered(); C
        Category of filtered modules with basis over Integer Ring
        sage: sorted(C.super_categories(), key=str)
        [Category of filtered modules over Integer Ring,
         Category of modules with basis over Integer Ring]
        sage: C is ModulesWithBasis(ZZ).Filtered()
        True

    TESTS::

        sage: C = ModulesWithBasis(ZZ).Filtered()
        sage: TestSuite(C).run()
        sage: C = ModulesWithBasis(QQ).Filtered()
        sage: TestSuite(C).run()
    """
    class ParentMethods:
        # TODO: which syntax do we prefer?
        # A.basis(degree = 3)
        # A.basis().subset(degree=3)

        # This is related to the following design question:
        # If F = (f_i)_{i\in I} is a family, should ``F.subset(degree = 3)``
        # be the elements of F of degree 3 or those whose index is of degree 3?

        @cached_method
        def basis(self, d=None):
            r"""
            Return the basis for (the ``d``-th homogeneous component
            of) ``self``.

            INPUT:

            - ``d`` -- (default: ``None``) nonnegative integer
              or ``None``

            OUTPUT:

            If ``d`` is ``None``, returns the basis of the module.
            Otherwise, returns the basis of the homogeneous component
            of degree ``d`` (i.e., the subfamily of the basis of the
            whole module which consists only of the basis vectors
            lying in `F_d \setminus \bigcup_{i<d} F_i`).

            The basis is always returned as a family.

            EXAMPLES::

                sage: A = ModulesWithBasis(ZZ).Filtered().example()
                sage: A.basis(4)
                Lazy family (Term map from Partitions to An example of a
                 filtered module with basis: the free module on partitions
                 over Integer Ring(i))_{i in Partitions of the integer 4}

            Without arguments, the full basis is returned::

                sage: A.basis()
                Lazy family (Term map from Partitions to An example of a
                 filtered module with basis: the free module on partitions
                 over Integer Ring(i))_{i in Partitions}
                sage: A.basis()
                Lazy family (Term map from Partitions to An example of a
                 filtered module with basis: the free module on partitions
                 over Integer Ring(i))_{i in Partitions}

            Checking this method on a filtered algebra. Note that this
            will typically raise a :exc:`NotImplementedError` when this
            feature is not implemented. ::

                sage: A = AlgebrasWithBasis(ZZ).Filtered().example()
                sage: A.basis(4)
                Traceback (most recent call last):
                ...
                NotImplementedError: infinite set

            Without arguments, the full basis is returned::

                sage: A.basis()
                Lazy family (Term map from Free abelian monoid indexed by
                 {'x', 'y', 'z'} to An example of a filtered algebra with
                 basis: the universal enveloping algebra of Lie algebra
                 of RR^3 with cross product over Integer Ring(i))_{i in
                 Free abelian monoid indexed by {'x', 'y', 'z'}}

            An example with a graded algebra::

                sage: E.<x,y> = ExteriorAlgebra(QQ)
                sage: E.basis()
                Lazy family (Term map from Subsets of {0,1} to
                 The exterior algebra of rank 2 over Rational Field(i))_{i in
                 Subsets of {0,1}}
            """
            if d is None:
                from sage.sets.family import Family
                return Family(self._indices, self.monomial)
            else:
                return self.homogeneous_component_basis(d)

        # TODO: Change `list(self._indices)` to `self._indices` and move
        #   this fallback to the category of finite-dimensional filtered
        #   modules with basis when it is implemented and the MRO issues
        #   are fixed (see trac #19397)
        def homogeneous_component_basis(self, d):
            """
            Return a basis for the ``d``-th homogeneous component of ``self``.

            EXAMPLES::

                sage: A = GradedModulesWithBasis(ZZ).example()
                sage: A.homogeneous_component_basis(4)
                Lazy family (Term map
                 from Partitions
                   to An example of a graded module with basis: the free module
                      on partitions over Integer Ring(i))_{i in Partitions of the integer 4}

                sage: # needs sage.modules
                sage: cat = GradedModulesWithBasis(ZZ)
                sage: C = CombinatorialFreeModule(ZZ, ['a', 'b'], category=cat)
                sage: C.degree_on_basis = lambda x: 1 if x == 'a' else 2
                sage: C.homogeneous_component_basis(1)
                Finite family {'a': B['a']}
                sage: C.homogeneous_component_basis(2)
                Finite family {'b': B['b']}
            """
            from sage.sets.family import Family
            try:
                S = self._indices.subset(size=d)
            except (AttributeError, ValueError, TypeError):
                S = [i for i in list(self._indices) if self.degree_on_basis(i) == d]
            return Family(S, self.monomial)

        def homogeneous_component(self, d):
            """
            Return the ``d``-th homogeneous component of ``self``.

            EXAMPLES::

                sage: A = GradedModulesWithBasis(ZZ).example()
                sage: A.homogeneous_component(4)
                Degree 4 homogeneous component of An example of a graded module
                 with basis: the free module on partitions over Integer Ring
            """
            from sage.categories.modules_with_basis import ModulesWithBasis
            from sage.categories.filtered_algebras import FilteredAlgebras
            if self.base_ring() in FilteredAlgebras:
                raise NotImplementedError("this is only a natural module over"
                                          " the degree 0 component of the filtered"
                                          " algebra and coordinate rings are not"
                                          " yet implemented for submodules")
            category = ModulesWithBasis(self.category().base_ring())
            M = self.submodule(self.homogeneous_component_basis(d),
                               category=category,
                               already_echelonized=True)
            M.rename("Degree {} homogeneous component of {}".format(d, self))
            return M

        def hilbert_series(self, prec=None):
            r"""
            Return the Hilbert series of ``self``.

            Let `R` be a commutative ring (with unit). Let
            `M = \bigcup_{n=0}^{\infty} M_n` be a filtered `R`-module.
            The *Hilbert series* of `M` is the formal power series

            .. MATH::

                H(t) = \sum_{n=0}^{\infty} \ell(M_n / M_{n-1}) t^n,

            where `\ell(N)` is the *length* of `N`, which is the
            longest chain of submodules (over `R`), and by convention
            `M_{-1} = \{0\}`. By the assumptions of the category,
            `M_n / M_{n-1}` is a free `R`-module, and so `\ell(M_n / M_{n-1})`
            is equal to the rank of `M_n / M_{n-1}`.

            INPUT:

            - ``prec`` -- (default: `\infty`) the precision

            OUTPUT:

            If the precision is finite, then this returns an element in the
            :class:`PowerSeriesRing` over ``ZZ``. Otherwise it returns an
            element in the :class:`LazyPowerSeriesRing` over ``ZZ``.

            EXAMPLES::

                sage: A = GradedModulesWithBasis(ZZ).example()
                sage: A.hilbert_series()
                1 + t + 2*t^2 + 3*t^3 + 5*t^4 + 7*t^5 + 11*t^6 + O(t^7)
                sage: A.hilbert_series(10)
                1 + t + 2*t^2 + 3*t^3 + 5*t^4 + 7*t^5 + 11*t^6 + 15*t^7 + 22*t^8 + 30*t^9 + O(t^10)
            """
            from sage.rings.integer_ring import ZZ
            if prec is None:
                from sage.rings.lazy_series_ring import LazyPowerSeriesRing
                R = LazyPowerSeriesRing(ZZ, 't')
                return R(lambda n: self.homogeneous_component_basis(n).cardinality())
            from sage.rings.power_series_ring import PowerSeriesRing
            R = PowerSeriesRing(ZZ, 't')
            elt = R([self.homogeneous_component_basis(n).cardinality() for n in range(prec)])
            return elt.O(prec)

        def graded_algebra(self):
            r"""
            Return the associated graded module to ``self``.

            See :class:`~sage.algebras.associated_graded.AssociatedGradedAlgebra`
            for the definition and the properties of this.

            If the filtered module ``self`` with basis is called `A`,
            then this method returns `\operatorname{gr} A`. The method
            :meth:`to_graded_conversion` returns the canonical
            `R`-module isomorphism `A \to \operatorname{gr} A` induced
            by the basis of `A`, and the method
            :meth:`from_graded_conversion` returns the inverse of this
            isomorphism. The method :meth:`projection` projects
            elements of `A` onto `\operatorname{gr} A` according to
            their place in the filtration on `A`.

            .. WARNING::

                When not overridden, this method returns the default
                implementation of an associated graded module --
                namely, ``AssociatedGradedAlgebra(self)``, where
                ``AssociatedGradedAlgebra`` is
                :class:`~sage.algebras.associated_graded.AssociatedGradedAlgebra`.
                But some instances of :class:`FilteredModulesWithBasis`
                override this method, as the associated graded module
                often is (isomorphic) to a simpler object (for instance,
                the associated graded module of a graded module can be
                identified with the graded module itself). Generic code
                that uses associated graded modules (such as the code
                of the :meth:`induced_graded_map` method below) should
                make sure to only communicate with them via the
                :meth:`to_graded_conversion`,
                :meth:`from_graded_conversion` and
                :meth:`projection` methods (in particular,
                do not expect there to be a conversion from ``self``
                to ``self.graded_algebra()``; this currently does not
                work for Clifford algebras). Similarly, when
                overriding :meth:`graded_algebra`, make sure to
                accordingly redefine these three methods, unless their
                definitions below still apply to your case (this will
                happen whenever the basis of your :meth:`graded_algebra`
                has the same indexing set as ``self``, and the partition
                of this indexing set according to degree is the same as
                for ``self``).

            EXAMPLES::

                sage: A = ModulesWithBasis(ZZ).Filtered().example()
                sage: A.graded_algebra()
                Graded Module of An example of a filtered module with basis:
                 the free module on partitions over Integer Ring
            """
            from sage.algebras.associated_graded import AssociatedGradedAlgebra
            return AssociatedGradedAlgebra(self)

        # Maps

        def to_graded_conversion(self):
            r"""
            Return the canonical `R`-module isomorphism
            `A \to \operatorname{gr} A` induced by the basis of `A`
            (where `A = ` ``self``).

            This is an isomorphism of `R`-modules. See
            the class documentation :class:`AssociatedGradedAlgebra`.

            .. SEEALSO::

                :meth:`from_graded_conversion`

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: A = Modules(QQ).WithBasis().Filtered().example()
                sage: p = -2 * A.an_element(); p
                -4*P[] - 4*P[1] - 6*P[2]
                sage: q = A.to_graded_conversion()(p); q
                -4*Bbar[[]] - 4*Bbar[[1]] - 6*Bbar[[2]]
                sage: q.parent() is A.graded_algebra()
                True
            """
            base_one = self.base_ring().one()
            return self.module_morphism(diagonal=lambda x: base_one,
                                        codomain=self.graded_algebra())

        def from_graded_conversion(self):
            r"""
            Return the inverse of the canonical `R`-module isomorphism
            `A \to \operatorname{gr} A` induced by the basis of `A`
            (where `A = ` ``self``). This inverse is an isomorphism
            `\operatorname{gr} A \to A`.

            This is an isomorphism of `R`-modules. See
            the class documentation :class:`AssociatedGradedAlgebra`.

            .. SEEALSO::

                :meth:`to_graded_conversion`

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: A = Modules(QQ).WithBasis().Filtered().example()
                sage: p = -2 * A.an_element(); p
                -4*P[] - 4*P[1] - 6*P[2]
                sage: q = A.to_graded_conversion()(p); q
                -4*Bbar[[]] - 4*Bbar[[1]] - 6*Bbar[[2]]
                sage: A.from_graded_conversion()(q) == p
                True
                sage: q.parent() is A.graded_algebra()
                True
            """
            base_one = self.base_ring().one()
            return self.graded_algebra().module_morphism(diagonal=lambda x: base_one,
                                                         codomain=self)

        def projection(self, i):
            r"""
            Return the `i`-th projection `p_i : F_i \to G_i` (in the
            notations of the class documentation
            :class:`AssociatedGradedAlgebra`, where `A = ` ``self``).

            This method actually does not return the map `p_i` itself,
            but an extension of `p_i` to the whole `R`-module `A`.
            This extension is the composition of the `R`-module
            isomorphism `A \to \operatorname{gr} A` with the canonical
            projection of the graded `R`-module `\operatorname{gr} A`
            onto its `i`-th graded component `G_i`. The codomain of
            this map is `\operatorname{gr} A`, although its actual
            image is `G_i`. The map `p_i` is obtained from this map
            by restricting its domain to `F_i` and its image to `G_i`.

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: A = Modules(ZZ).WithBasis().Filtered().example()
                sage: p = -2 * A.an_element(); p
                -4*P[] - 4*P[1] - 6*P[2]
                sage: q = A.projection(2)(p); q
                -6*Bbar[[2]]
                sage: q.parent() is A.graded_algebra()
                True
                sage: A.projection(3)(p)
                0
            """
            base_zero = self.base_ring().zero()
            base_one = self.base_ring().one()
            grA = self.graded_algebra()
            proj = lambda x: (base_one if self.degree_on_basis(x) == i
                              else base_zero)
            return self.module_morphism(diagonal=proj, codomain=grA)

        def induced_graded_map(self, other, f):
            r"""
            Return the graded linear map between the associated graded
            modules of ``self`` and ``other`` canonically induced by
            the filtration-preserving map ``f : self -> other``.

            Let `A` and `B` be two filtered modules with basis, and let
            `(F_i)_{i \in I}` and `(G_i)_{i \in I}` be their
            filtrations. Let `f : A \to B` be a linear map which
            preserves the filtration (i.e., satisfies `f(F_i) \subseteq
            G_i` for all `i \in I`). Then, there is a canonically
            defined graded linear map
            `\operatorname{gr} f : \operatorname{gr} A \to
            \operatorname{gr} B` which satisfies

            .. MATH::

                (\operatorname{gr} f) (p_i(a)) = p_i(f(a))
                \qquad \text{for all } i \in I \text{ and } a \in F_i ,

            where the `p_i` on the left hand side is the canonical
            projection from `F_i` onto the `i`-th graded component
            of `\operatorname{gr} A`, while the `p_i` on the right
            hand side is the canonical projection from `G_i` onto
            the `i`-th graded component of `\operatorname{gr} B`.

            INPUT:

            - ``other`` -- a filtered algebra with basis

            - ``f`` -- a filtration-preserving linear map from ``self``
              to ``other`` (can be given as a morphism or as a function)

            OUTPUT: the graded linear map `\operatorname{gr} f`

            EXAMPLES:

            **Example 1.**

            We start with the free `\QQ`-module with basis the set of all
            partitions::

                sage: A = Modules(QQ).WithBasis().Filtered().example(); A               # needs sage.combinat sage.modules
                An example of a filtered module with basis: the free module
                 on partitions over Rational Field
                sage: M = A.indices(); M                                                # needs sage.combinat sage.modules
                Partitions
                sage: p1, p2, p21, p321 = [A.basis()[Partition(i)]                      # needs sage.combinat sage.modules
                ....:                      for i in [[1], [2], [2,1], [3,2,1]]]

            Let us define a map from ``A`` to itself which acts on the
            basis by sending every partition `\lambda` to the sum of
            the conjugates of all partitions `\mu` for which
            `\lambda / \mu` is a horizontal strip::

                sage: # needs sage.combinat sage.modules
                sage: def map_on_basis(lam):
                ....:     def mus(k):
                ....:         return lam.remove_horizontal_border_strip(k)
                ....:     return A.sum_of_monomials([Partition(mu).conjugate()
                ....:                                for k in range(sum(lam) + 1)
                ....:                                for mu in mus(k)])
                sage: f = A.module_morphism(on_basis=map_on_basis,
                ....:                       codomain=A)
                sage: f(p1)
                P[] + P[1]
                sage: f(p2)
                P[] + P[1] + P[1, 1]
                sage: f(p21)
                P[1] + P[1, 1] + P[2] + P[2, 1]
                sage: f(p21 - p1)
                -P[] + P[1, 1] + P[2] + P[2, 1]
                sage: f(p321)
                P[2, 1] + P[2, 1, 1] + P[2, 2] + P[2, 2, 1]
                 + P[3, 1] + P[3, 1, 1] + P[3, 2] + P[3, 2, 1]

            We now compute `\operatorname{gr} f` ::

                sage: # needs sage.combinat sage.modules
                sage: grA = A.graded_algebra(); grA
                Graded Module of An example of a filtered module with basis:
                 the free module on partitions over Rational Field
                sage: pp1, pp2, pp21, pp321 = [A.to_graded_conversion()(i)
                ....:                          for i in [p1, p2, p21, p321]]
                sage: pp2 + 4 * pp21
                Bbar[[2]] + 4*Bbar[[2, 1]]
                sage: grf = A.induced_graded_map(A, f); grf
                Generic endomorphism of Graded Module of
                 An example of a filtered module with basis:
                  the free module on partitions over Rational Field
                sage: grf(pp1)
                Bbar[[1]]
                sage: grf(pp2 + 4 * pp21)
                Bbar[[1, 1]] + 4*Bbar[[2, 1]]

            **Example 2.**

            We shall now construct `\operatorname{gr} f` for a
            different map `f` out of the same ``A``; the new map
            `f` will lead into a graded algebra already, namely into
            the algebra of symmetric functions::

                sage: # needs sage.combinat sage.modules
                sage: h = SymmetricFunctions(QQ).h()
                sage: def map_on_basis(lam):  # redefining map_on_basis
                ....:     def mus(k):
                ....:         return lam.remove_horizontal_border_strip(k)
                ....:     return h.sum_of_monomials([Partition(mu).conjugate()
                ....:                                for k in range(sum(lam) + 1)
                ....:                                for mu in mus(k)])
                sage: f = A.module_morphism(on_basis=map_on_basis,
                ....:                       codomain=h)  # redefining f
                sage: f(p1)
                h[] + h[1]
                sage: f(p2)
                h[] + h[1] + h[1, 1]
                sage: f(A.zero())
                0
                sage: f(p2 - 3*p1)
                -2*h[] - 2*h[1] + h[1, 1]

            The algebra ``h`` of symmetric functions in the `h`-basis
            is already graded, so its associated graded algebra is
            implemented as itself::

                sage: # needs sage.combinat sage.modules
                sage: grh = h.graded_algebra(); grh is h
                True
                sage: grf = A.induced_graded_map(h, f); grf
                Generic morphism:
                  From: Graded Module of An example of a filtered
                   module with basis: the free module on partitions
                   over Rational Field
                  To:   Symmetric Functions over Rational Field
                   in the homogeneous basis
                sage: grf(pp1)
                h[1]
                sage: grf(pp2)
                h[1, 1]
                sage: grf(pp321)
                h[3, 2, 1]
                sage: grf(pp2 - 3*pp1)
                -3*h[1] + h[1, 1]
                sage: grf(pp21)
                h[2, 1]
                sage: grf(grA.zero())
                0

            **Example 3.**

            After having had a graded module as the codomain, let us try to
            have one as the domain instead. Our new ``f`` will go from ``h``
            to ``A``::

                sage: # needs sage.combinat sage.modules
                sage: def map_on_basis(lam):  # redefining map_on_basis
                ....:     def mus(k):
                ....:         return lam.remove_horizontal_border_strip(k)
                ....:     return A.sum_of_monomials([Partition(mu).conjugate()
                ....:                                for k in range(sum(lam) + 1)
                ....:                                for mu in mus(k)])
                sage: f = h.module_morphism(on_basis=map_on_basis,
                ....:                       codomain=A)  # redefining f
                sage: f(h[1])
                P[] + P[1]
                sage: f(h[2])
                P[] + P[1] + P[1, 1]
                sage: f(h[1, 1])
                P[1] + P[2]
                sage: f(h[2, 2])
                P[1, 1] + P[2, 1] + P[2, 2]
                sage: f(h[3, 2, 1])
                P[2, 1] + P[2, 1, 1] + P[2, 2] + P[2, 2, 1]
                 + P[3, 1] + P[3, 1, 1] + P[3, 2] + P[3, 2, 1]
                sage: f(h.one())
                P[]
                sage: grf = h.induced_graded_map(A, f); grf
                Generic morphism:
                  From: Symmetric Functions over Rational Field
                   in the homogeneous basis
                  To:   Graded Module of An example of a filtered
                   module with basis: the free module on partitions
                   over Rational Field
                sage: grf(h[1])
                Bbar[[1]]
                sage: grf(h[2])
                Bbar[[1, 1]]
                sage: grf(h[1, 1])
                Bbar[[2]]
                sage: grf(h[2, 2])
                Bbar[[2, 2]]
                sage: grf(h[3, 2, 1])
                Bbar[[3, 2, 1]]
                sage: grf(h.one())
                Bbar[[]]

            **Example 4.**

            The construct `\operatorname{gr} f` also makes sense when `f`
            is a filtration-preserving map between graded modules. ::

                sage: # needs sage.combinat sage.modules
                sage: def map_on_basis(lam):  # redefining map_on_basis
                ....:     def mus(k):
                ....:         return lam.remove_horizontal_border_strip(k)
                ....:     return h.sum_of_monomials([Partition(mu).conjugate()
                ....:                                for k in range(sum(lam) + 1)
                ....:                                for mu in mus(k)])
                sage: f = h.module_morphism(on_basis=map_on_basis,
                ....:                       codomain=h)  # redefining f
                sage: f(h[1])
                h[] + h[1]
                sage: f(h[2])
                h[] + h[1] + h[1, 1]
                sage: f(h[1, 1])
                h[1] + h[2]
                sage: f(h[2, 1])
                h[1] + h[1, 1] + h[2] + h[2, 1]
                sage: f(h.one())
                h[]
                sage: grf = h.induced_graded_map(h, f); grf
                Generic endomorphism of
                 Symmetric Functions over Rational Field in the homogeneous basis
                sage: grf(h[1])
                h[1]
                sage: grf(h[2])
                h[1, 1]
                sage: grf(h[1, 1])
                h[2]
                sage: grf(h[2, 1])
                h[2, 1]
                sage: grf(h.one())
                h[]
            """
            grA = self.graded_algebra()
            grB = other.graded_algebra()
            from sage.categories.graded_modules_with_basis import GradedModulesWithBasis
            cat = GradedModulesWithBasis(self.base_ring())
            from_gr = self.from_graded_conversion()

            def on_basis(m):
                i = grA.degree_on_basis(m)
                lifted_img_of_m = f(from_gr(grA.monomial(m)))
                return other.projection(i)(lifted_img_of_m)
            return grA.module_morphism(on_basis=on_basis,
                                       codomain=grB, category=cat)
            # If we could assume that the projection of the basis
            # element of ``self`` indexed by an index ``m`` is the
            # basis element of ``grA`` indexed by ``m``, then this
            # could go faster:
            #
            # def on_basis(m):
            #     i = grA.degree_on_basis(m)
            #     return grB.projection(i)(f(self.monomial(m)))
            # return grA.module_morphism(on_basis=on_basis,
            #                            codomain=grB, category=cat)
            #
            # But this assumption might come back to bite us in the
            # ass one day. What do you think?

    class ElementMethods:

        def is_homogeneous(self):
            r"""
            Return whether the element ``self`` is homogeneous.

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: A = ModulesWithBasis(ZZ).Filtered().example()
                sage: x = A(Partition((3,2,1)))
                sage: y = A(Partition((4,4,1)))
                sage: z = A(Partition((2,2,2)))
                sage: (3*x).is_homogeneous()
                True
                sage: (x - y).is_homogeneous()
                False
                sage: (x+2*z).is_homogeneous()
                True

            Here is an example with a graded algebra::

                sage: # needs sage.combinat sage.modules
                sage: S = NonCommutativeSymmetricFunctions(QQ).S()
                sage: (x, y) = (S[2], S[3])
                sage: (3*x).is_homogeneous()
                True
                sage: (x^3 - y^2).is_homogeneous()
                True
                sage: ((x + y)^2).is_homogeneous()
                False

            Let us now test a filtered algebra (but remember that the
            notion of homogeneity now depends on the choice of a
            basis, or at least on a definition of homogeneous
            components)::

                sage: A = AlgebrasWithBasis(QQ).Filtered().example()
                sage: x,y,z = A.algebra_generators()
                sage: (x*y).is_homogeneous()
                True
                sage: (y*x).is_homogeneous()
                False
                sage: A.one().is_homogeneous()
                True
                sage: A.zero().is_homogeneous()
                True
                sage: (A.one()+x).is_homogeneous()
                False
            """
            degree_on_basis = self.parent().degree_on_basis
            degree = None
            for m in self.support():
                if degree is None:
                    degree = degree_on_basis(m)
                else:
                    if degree != degree_on_basis(m):
                        return False
            return True

        @abstract_method(optional=True)
        def degree_on_basis(self, m):
            r"""
            Return the degree of the basis element indexed by ``m``
            in ``self``.

            EXAMPLES::

                sage: A = GradedModulesWithBasis(QQ).example()                          # needs sage.combinat sage.modules
                sage: A.degree_on_basis(Partition((2,1)))                               # needs sage.combinat sage.modules
                3
                sage: A.degree_on_basis(Partition((4,2,1,1,1,1)))                       # needs sage.combinat sage.modules
                10
            """

        def homogeneous_degree(self):
            r"""
            The degree of a nonzero homogeneous element ``self`` in the
            filtered module.

            .. NOTE::

                This raises an error if the element is not homogeneous.
                To compute the maximum of the degrees of the homogeneous
                summands of a (not necessarily homogeneous) element, use
                :meth:`maximal_degree` instead.

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: A = ModulesWithBasis(ZZ).Filtered().example()
                sage: x = A(Partition((3,2,1)))
                sage: y = A(Partition((4,4,1)))
                sage: z = A(Partition((2,2,2)))
                sage: x.degree()
                6
                sage: (x + 2*z).degree()
                6
                sage: (y - x).degree()
                Traceback (most recent call last):
                ...
                ValueError: element is not homogeneous

            An example in a graded algebra::

                sage: # needs sage.combinat sage.modules
                sage: S = NonCommutativeSymmetricFunctions(QQ).S()
                sage: (x, y) = (S[2], S[3])
                sage: x.homogeneous_degree()
                2
                sage: (x^3 + 4*y^2).homogeneous_degree()
                6
                sage: ((1 + x)^3).homogeneous_degree()
                Traceback (most recent call last):
                ...
                ValueError: element is not homogeneous

            Let us now test a filtered algebra (but remember that the
            notion of homogeneity now depends on the choice of a
            basis)::

                sage: A = AlgebrasWithBasis(QQ).Filtered().example()
                sage: x,y,z = A.algebra_generators()
                sage: (x*y).homogeneous_degree()
                2
                sage: (y*x).homogeneous_degree()
                Traceback (most recent call last):
                ...
                ValueError: element is not homogeneous
                sage: A.one().homogeneous_degree()
                0

            TESTS::

                sage: S = NonCommutativeSymmetricFunctions(QQ).S()                      # needs sage.combinat sage.modules
                sage: S.zero().degree()                                                 # needs sage.combinat sage.modules
                Traceback (most recent call last):
                ...
                ValueError: the zero element does not have a well-defined degree
            """
            if not self.support():
                raise ValueError("the zero element does not have a well-defined degree")
            if not self.is_homogeneous():
                raise ValueError("element is not homogeneous")
            return self.parent().degree_on_basis(next(iter(self.support())))

        # default choice for degree; will be overridden as necessary
        degree = homogeneous_degree

        def maximal_degree(self):
            """
            The maximum of the degrees of the homogeneous components
            of ``self``.

            This is also the smallest `i` such that ``self`` belongs
            to `F_i`. Hence, it does not depend on the basis of the
            parent of ``self``.

            .. SEEALSO:: :meth:`homogeneous_degree`

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: A = ModulesWithBasis(ZZ).Filtered().example()
                sage: x = A(Partition((3,2,1)))
                sage: y = A(Partition((4,4,1)))
                sage: z = A(Partition((2,2,2)))
                sage: x.maximal_degree()
                6
                sage: (x + 2*z).maximal_degree()
                6
                sage: (y - x).maximal_degree()
                9
                sage: (3*z).maximal_degree()
                6

            Now, we test this on a graded algebra::

                sage: # needs sage.combinat sage.modules
                sage: S = NonCommutativeSymmetricFunctions(QQ).S()
                sage: (x, y) = (S[2], S[3])
                sage: x.maximal_degree()
                2
                sage: (x^3 + 4*y^2).maximal_degree()
                6
                sage: ((1 + x)^3).maximal_degree()
                6

            Let us now test a filtered algebra::

                sage: # needs sage.combinat sage.modules
                sage: A = AlgebrasWithBasis(QQ).Filtered().example()
                sage: x,y,z = A.algebra_generators()
                sage: (x*y).maximal_degree()
                2
                sage: (y*x).maximal_degree()
                2
                sage: A.one().maximal_degree()
                0
                sage: A.zero().maximal_degree()
                Traceback (most recent call last):
                ...
                ValueError: the zero element does not have a well-defined degree
                sage: (A.one()+x).maximal_degree()
                1

            TESTS::

                sage: S = NonCommutativeSymmetricFunctions(QQ).S()                      # needs sage.combinat sage.modules
                sage: S.zero().degree()                                                 # needs sage.combinat sage.modules
                Traceback (most recent call last):
                ...
                ValueError: the zero element does not have a well-defined degree
            """
            if self.is_zero():
                raise ValueError("the zero element does not have a well-defined degree")
            degree_on_basis = self.parent().degree_on_basis
            return max(degree_on_basis(m) for m in self.support())

        def homogeneous_component(self, n):
            r"""
            Return the homogeneous component of degree ``n`` of the
            element ``self``.

            Let `m` be an element of a filtered `R`-module `M` with
            basis. Then, `m` can be uniquely written in the form
            `m = \sum_{i \in I} m_i`, where each `m_i` is a
            homogeneous element of degree `i`. For `n \in I`, we
            define the homogeneous component of degree `n` of the
            element `m` to be `m_n`.

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: A = ModulesWithBasis(ZZ).Filtered().example()
                sage: x = A.an_element(); x
                2*P[] + 2*P[1] + 3*P[2]
                sage: x.homogeneous_component(-1)
                0
                sage: x.homogeneous_component(0)
                2*P[]
                sage: x.homogeneous_component(1)
                2*P[1]
                sage: x.homogeneous_component(2)
                3*P[2]
                sage: x.homogeneous_component(3)
                0

                sage: # needs sage.combinat sage.modules
                sage: A = ModulesWithBasis(ZZ).Graded().example()
                sage: x = A.an_element(); x
                2*P[] + 2*P[1] + 3*P[2]
                sage: x.homogeneous_component(-1)
                0
                sage: x.homogeneous_component(0)
                2*P[]
                sage: x.homogeneous_component(1)
                2*P[1]
                sage: x.homogeneous_component(2)
                3*P[2]
                sage: x.homogeneous_component(3)
                0

                sage: A = AlgebrasWithBasis(ZZ).Filtered().example()
                sage: G = A.algebra_generators()
                sage: g = A.an_element() - 2 * G['x'] * G['y']; g
                U['x']^2*U['y']^2*U['z']^3 - 2*U['x']*U['y']
                 + 2*U['x'] + 3*U['y'] + 1
                sage: g.homogeneous_component(-1)
                0
                sage: g.homogeneous_component(0)
                1
                sage: g.homogeneous_component(2)
                -2*U['x']*U['y']
                sage: g.homogeneous_component(5)
                0
                sage: g.homogeneous_component(7)
                U['x']^2*U['y']^2*U['z']^3
                sage: g.homogeneous_component(8)
                0

            TESTS:

            Check that this really returns ``A.zero()`` and not a plain ``0``::

                sage: # needs sage.combinat sage.modules
                sage: A = ModulesWithBasis(ZZ).Filtered().example()
                sage: x = A.an_element()
                sage: x.homogeneous_component(3).parent() is A
                True
            """
            degree_on_basis = self.parent().degree_on_basis
            return self.parent().sum_of_terms((i, c)
                                              for (i, c) in self
                                              if degree_on_basis(i) == n)

        def truncate(self, n):
            """
            Return the sum of the homogeneous components of degree
            strictly less than ``n`` of ``self``.

            See :meth:`homogeneous_component` for the notion of a
            homogeneous component.

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: A = ModulesWithBasis(ZZ).Filtered().example()
                sage: x = A.an_element(); x
                2*P[] + 2*P[1] + 3*P[2]
                sage: x.truncate(0)
                0
                sage: x.truncate(1)
                2*P[]
                sage: x.truncate(2)
                2*P[] + 2*P[1]
                sage: x.truncate(3)
                2*P[] + 2*P[1] + 3*P[2]

                sage: # needs sage.combinat sage.modules
                sage: A = ModulesWithBasis(ZZ).Graded().example()
                sage: x = A.an_element(); x
                2*P[] + 2*P[1] + 3*P[2]
                sage: x.truncate(0)
                0
                sage: x.truncate(1)
                2*P[]
                sage: x.truncate(2)
                2*P[] + 2*P[1]
                sage: x.truncate(3)
                2*P[] + 2*P[1] + 3*P[2]

                sage: A = AlgebrasWithBasis(ZZ).Filtered().example()
                sage: G = A.algebra_generators()
                sage: g = A.an_element() - 2 * G['x'] * G['y']; g
                U['x']^2*U['y']^2*U['z']^3 - 2*U['x']*U['y']
                 + 2*U['x'] + 3*U['y'] + 1
                sage: g.truncate(-1)
                0
                sage: g.truncate(0)
                0
                sage: g.truncate(2)
                2*U['x'] + 3*U['y'] + 1
                sage: g.truncate(3)
                -2*U['x']*U['y'] + 2*U['x'] + 3*U['y'] + 1
                sage: g.truncate(5)
                -2*U['x']*U['y'] + 2*U['x'] + 3*U['y'] + 1
                sage: g.truncate(7)
                -2*U['x']*U['y'] + 2*U['x'] + 3*U['y'] + 1
                sage: g.truncate(8)
                U['x']^2*U['y']^2*U['z']^3 - 2*U['x']*U['y']
                 + 2*U['x'] + 3*U['y'] + 1

            TESTS:

            Check that this really return ``A.zero()`` and not a plain ``0``::

                sage: # needs sage.combinat sage.modules
                sage: A = ModulesWithBasis(ZZ).Filtered().example()
                sage: x = A.an_element()
                sage: x.truncate(0).parent() is A
                True
            """
            degree_on_basis = self.parent().degree_on_basis
            return self.parent().sum_of_terms((i, c) for (i, c) in self
                                              if degree_on_basis(i) < n)

    class Subobjects(SubobjectsCategory):
        class ParentMethods:
            def degree_on_basis(self, m):
                r"""
                Return the degree of the basis element indexed by ``m``
                in ``self``.

                EXAMPLES::

                    sage: # needs sage.modules
                    sage: E.<x,y,z> = ExteriorAlgebra(QQ)
                    sage: S = E.submodule([x + y, x*y - y*z, y])
                    sage: B = S.basis()
                    sage: [B[0].lift(), B[1].lift(), B[2].lift()]
                    [x, y, x*y - y*z]
                    sage: S.degree_on_basis(0)
                    1
                    sage: S.degree_on_basis(1)
                    1
                    sage: S.degree_on_basis(2)
                    2
                """
                return self.basis()[m].lift().degree()

        class ElementMethods:
            def degree(self):
                r"""
                Return the degree of ``self``.

                EXAMPLES::

                    sage: # needs sage.modules
                    sage: E.<x,y,z> = ExteriorAlgebra(QQ)
                    sage: S = E.submodule([x + y, x*y - y*z, y])
                    sage: B = S.basis()
                    sage: [B[0].lift(), B[1].lift(), B[2].lift()]
                    [x, y, x*y - y*z]
                    sage: B[0].degree()
                    1
                    sage: B[1].degree()
                    1
                    sage: (B[0] + 3*B[1]).degree()
                    1

                The degree of inhomogeneous elements is not defined
                (following the behavior of the exterior algebra)::

                    sage: (B[0] + B[2]).degree()                                        # needs sage.modules
                    Traceback (most recent call last):
                    ...
                    ValueError: element is not homogeneous

                We can still get the maximal degree::

                    sage: (B[0] + B[2]).maximal_degree()                                # needs sage.modules
                    2
                """
                return self.lift().degree()

            def maximal_degree(self):
                r"""
                The maximum of the degrees of the homogeneous components
                of ``self``.

                This is also the smallest `i` such that ``self`` belongs
                to `F_i`. Hence, it does not depend on the basis of the
                parent of ``self``.

                .. SEEALSO:: :meth:`homogeneous_degree`

                EXAMPLES::

                    sage: # needs sage.modules
                    sage: E.<x,y,z> = ExteriorAlgebra(QQ)
                    sage: F = E.submodule([x + 1, x*y - 1])
                    sage: B = F.basis()
                    sage: [B[0].lift(), B[1].lift()]
                    [-x*y + 1, x*y + x]
                    sage: B[0].maximal_degree()
                    2
                    sage: B[1].maximal_degree()
                    2
                """
                return self.lift().maximal_degree()

    class FiniteDimensional(CategoryWithAxiom_over_base_ring):
        class ParentMethods:
            def hilbert_series(self, prec=None):
                r"""
                Return the Hilbert series of ``self`` as a polynomial.

                Let `R` be a commutative ring (with unit). Let
                `M = \bigcup_{n=0}^{\infty} M_n` be a filtered `R`-module.
                The *Hilbert series* of `M` is the formal power series

                .. MATH::

                    H(t) = \sum_{n=0}^{\infty} \ell(M_n / M_{n-1}) t^n,

                where `\ell(N)` is the *length* of `N`, which is the
                longest chain of submodules (over `R`), and by convention
                `M_{-1} = \{0\}`. By the assumptions of the category,
                `M_n / M_{n-1}` is a free `R`-module, and so
                `\ell(M_n / M_{n-1})` is equal to the rank of `M_n / M_{n-1}`.

                EXAMPLES::

                    sage: OS = hyperplane_arrangements.braid(3).orlik_solomon_algebra(QQ)
                    sage: OS.hilbert_series()
                    2*t^2 + 3*t + 1

                    sage: OS = matroids.Uniform(5, 3).orlik_solomon_algebra(ZZ)
                    sage: OS.hilbert_series()
                    t^3 + 3*t^2 + 3*t + 1

                    sage: OS = matroids.PG(2, 3).orlik_solomon_algebra(ZZ['x','y'])
                    sage: OS.hilbert_series()
                    27*t^3 + 39*t^2 + 13*t + 1
                """
                from collections import defaultdict
                from sage.rings.integer_ring import ZZ
                from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
                PR = PolynomialRing(ZZ, 't')
                dims = defaultdict(ZZ)
                for b in self.basis():
                    dims[b.homogeneous_degree()] += 1
                return PR(dims)



# sage_setup: distribution = sagemath-categories
r"""
Examples of semirings
"""
# ****************************************************************************
#  Copyright (C) 2024 F. Chapoton <chapoton unistra.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.semirings import Semirings
from sage.structure.element import Element
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation

# semantic :
# 0 => 0, empty;
# 1 => 1, unique;
# 2 => at least 2
_ADD = [[0, 1, 2], [1, 2, 2], [2, 2, 2]]
_PROD = [[0, 0, 0], [0, 1, 2], [0, 2, 2]]


class Ternary(Element):
    """
    Elements of the ternary-logic ring.

    The semantic is as follows:

    - 0 -- the integer 0
    - 1 -- the integer 1
    - 2 -- some integer greater than 1

    An alternative semantic is:

    - 0 -- an empty set
    - 1 -- a connected set
    - 2 -- a disconnected set

    The same semantic works for graphs instead of sets.
    """
    def __init__(self, parent, n):
        """
        Initialize one element.

        TESTS::

            sage: from sage.categories.examples.semirings import TernaryLogic
            sage: S = TernaryLogic()
            sage: S(4)
            Traceback (most recent call last):
            ...
            ValueError: input not in (0,1,2)
        """
        if n not in [0, 1, 2]:
            raise ValueError("input not in (0,1,2)")
        self._n = n
        Element.__init__(self, parent)

    def _repr_(self):
        """
        Return the string representation.

        TESTS::

            sage: from sage.categories.examples.semirings import TernaryLogic
            sage: S = TernaryLogic()
            sage: [S(i) for i in range(3)]
            [0, 1, many]
        """
        return ["0", "1", "many"][self._n]

    def __eq__(self, other):
        """
        Test for equality.

        TESTS::

            sage: from sage.categories.examples.semirings import TernaryLogic
            sage: S = TernaryLogic()
            sage: S(1) == S(2)
            False
            sage: S(0) == 3
            False
        """
        if not isinstance(other, Ternary):
            return False
        return self._n == other._n

    def __ne__(self, other):
        """
        Test for non-equality.

        TESTS::

            sage: from sage.categories.examples.semirings import TernaryLogic
            sage: S = TernaryLogic()
            sage: S(1) != S(2)
            True
        """
        return not (self == other)


class TernaryLogic(UniqueRepresentation, Parent):
    r"""
    An example of a semiring.

    This class illustrates a minimal implementation of a semiring.

    EXAMPLES::

        sage: S = Semirings().example(); S
        An example of a semiring: the ternary-logic semiring

    This is the semiring that contains 3 objects::

        sage: S.some_elements()
        [0, 1, many]

    The product rule is as expected::

        sage: S(1) * S(1)
        1
        sage: S(1) + S(1)
        many

    TESTS::

        sage: TestSuite(S).run(verbose=True)
        running ._test_additive_associativity() . . . pass
        running ._test_an_element() . . . pass
        running ._test_associativity() . . . pass
        running ._test_cardinality() . . . pass
        running ._test_category() . . . pass
        running ._test_construction() . . . pass
        running ._test_distributivity() . . . pass
        running ._test_elements() . . .
          Running the test suite of self.an_element()
          running ._test_category() . . . pass
          running ._test_eq() . . . pass
          running ._test_new() . . . pass
          running ._test_nonzero_equal() . . . pass
          running ._test_not_implemented_methods() . . . pass
          running ._test_pickling() . . . pass
          pass
        running ._test_elements_eq_reflexive() . . . pass
        running ._test_elements_eq_symmetric() . . . pass
        running ._test_elements_eq_transitive() . . . pass
        running ._test_elements_neq() . . . pass
        running ._test_eq() . . . pass
        running ._test_new() . . . pass
        running ._test_not_implemented_methods() . . . pass
        running ._test_one() . . . pass
        running ._test_pickling() . . . pass
        running ._test_prod() . . . pass
        running ._test_some_elements() . . . pass
        running ._test_zero() . . . pass
    """
    def __init__(self):
        r"""
        The ternary-logic semiring.

        EXAMPLES::

            sage: S = Semirings().example(); S
            An example of a semiring: the ternary-logic semiring
        """
        Parent.__init__(self, category=Semirings())

    def _repr_(self):
        r"""
        Return the string representation.

        EXAMPLES::

            sage: Semirings().example()._repr_()
            'An example of a semiring: the ternary-logic semiring'
        """
        return "An example of a semiring: the ternary-logic semiring"

    def summation(self, x, y):
        r"""
        Return the sum of ``x`` and ``y`` in the semiring as per
        :meth:`Semirings.ParentMethods.summation`.

        EXAMPLES::

            sage: S = Semirings().example()
            sage: S(1) + S(1)
            many
        """
        assert x in self
        assert y in self
        return self(_ADD[x._n][y._n])

    def one(self):
        """
        Return the unit of ``self``.

        EXAMPLES::

            sage: S = Semirings().example()
            sage: S.one()
            1
        """
        return self(1)

    def product(self, x, y):
        r"""
        Return the product of ``x`` and ``y`` in the semiring as per
        :meth:`Semirings.ParentMethods.product`.

        EXAMPLES::

            sage: S = Semirings().example()
            sage: S(1) * S(2)
            many
        """
        assert x in self
        assert y in self
        return self(_PROD[x._n][y._n])

    def an_element(self):
        r"""
        Return an element of the semiring.

        EXAMPLES::

            sage: Semirings().example().an_element()
            many
        """
        return self(2)

    def some_elements(self):
        r"""
        Return a list of some elements of the semiring.

        EXAMPLES::

            sage: Semirings().example().some_elements()
            [0, 1, many]
        """
        return [self(i) for i in [0, 1, 2]]

    Element = Ternary


Example = TernaryLogic



# sage_setup: distribution = sagemath-categories
r"""
Euclidean domains

AUTHORS:

- Teresa Gomez-Diaz (2008): initial version

- Julian Rueth (2013-09-13): added euclidean degree, quotient remainder, and
  their tests
"""
# ****************************************************************************
#  Copyright (C) 2008 Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#                2013 Julian Rueth <julian.rueth@fsfe.org>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.categories.category_singleton import Category_singleton
from sage.categories.principal_ideal_domains import PrincipalIdealDomains
from sage.misc.abstract_method import abstract_method
from sage.structure.element import coerce_binop
from sage.structure.sequence import Sequence


class EuclideanDomains(Category_singleton):
    """
    The category of constructive euclidean domains, i.e., one can divide
    producing a quotient and a remainder where the remainder is either zero or
    its :meth:`ElementMethods.euclidean_degree` is smaller than the divisor.

    EXAMPLES::

      sage: EuclideanDomains()
      Category of euclidean domains
      sage: EuclideanDomains().super_categories()
      [Category of principal ideal domains]

    TESTS::

        sage: TestSuite(EuclideanDomains()).run()
    """
    def super_categories(self):
        """
        EXAMPLES::

            sage: EuclideanDomains().super_categories()
            [Category of principal ideal domains]
        """
        return [PrincipalIdealDomains()]

    class ParentMethods:
        def is_euclidean_domain(self):
            """
            Return True, since this in an object of the category of Euclidean domains.

            EXAMPLES::

                sage: Parent(QQ,category=EuclideanDomains()).is_euclidean_domain()
                True
            """
            return True

        def gcd_free_basis(self, elts):
            r"""
            Compute a set of coprime elements that can be used to express the
            elements of ``elts``.

            INPUT:

            - ``elts`` -- a sequence of elements of ``self``

            OUTPUT:

            A GCD-free basis (also called a coprime base) of ``elts``; that is,
            a set of pairwise relatively prime elements of ``self`` such that
            any element of ``elts`` can be written as a product of elements of
            the set.

            ALGORITHM:

            Naive implementation of the algorithm described in Section 4.8 of
            Bach & Shallit [BS1996]_.

            EXAMPLES::

                sage: ZZ.gcd_free_basis([1])
                []
                sage: ZZ.gcd_free_basis([4, 30, 14, 49])
                [2, 15, 7]

                sage: Pol.<x> = QQ[]
                sage: sorted(Pol.gcd_free_basis([
                ....:     (x+1)^3*(x+2)^3*(x+3), (x+1)*(x+2)*(x+3),
                ....:     (x+1)*(x+2)*(x+4)]))
                [x + 3, x + 4, x^2 + 3*x + 2]

            TESTS::

                sage: R.<x> = QQ[]
                sage: QQ.gcd_free_basis([x+1,x+2])
                Traceback (most recent call last):
                ...
                TypeError: unable to convert x + 1 to an element of Rational Field
            """
            def refine(a, b):
                g = a.gcd(b)
                if g.is_unit():
                    return (a, set(), b)
                l1, s1, r1 = refine(a//g, g)
                l2, s2, r2 = refine(r1, b//g)
                s1.update(s2)
                s1.add(l2)
                return (l1, s1, r2)
            elts = Sequence(elts, universe=self)
            res = set()
            if len(elts) == 1:
                res.update(elts)
            else:
                r = elts[-1]
                for t in self.gcd_free_basis(elts[:-1]):
                    l, s, r = refine(t, r)
                    res.update(s)
                    res.add(l)
                res.add(r)
            units = [x for x in res if x.is_unit()]
            res.difference_update(units)
            return Sequence(res, universe=self, check=False)

        def _test_euclidean_degree(self, **options):
            r"""
            Test that the assumptions on a Euclidean degree are met.

            EXAMPLES::

                sage: R.<x> = QQ[]
                sage: R._test_euclidean_degree()

            .. SEEALSO::

                :meth:`_test_quo_rem`
            """
            tester = self._tester(**options)
            S = [s for s in tester.some_elements() if not s.is_zero()]

            min_degree = self.one().euclidean_degree()

            from sage.rings.semirings.non_negative_integer_semiring import NN
            for a in S:
                tester.assertIn(a.euclidean_degree(), NN)
                tester.assertGreaterEqual(a.euclidean_degree(), min_degree)
                tester.assertEqual(a.euclidean_degree() == min_degree, a.is_unit())

            from sage.misc.misc import some_tuples
            for a,b in some_tuples(S, 2, tester._max_runs):
                p = a * b
                # For rings which are not exact, we might get something that
                #   acts like a zero divisor.
                # Therefore we skip the product if it evaluates to zero.
                # Let the category of Domains handle the test for zero divisors.
                if p.is_zero():
                    continue
                tester.assertLessEqual(a.euclidean_degree(), p.euclidean_degree())

        def _test_quo_rem(self, **options):
            r"""
            Test that the assumptions on a quotient with remainder of an
            euclidean domain are met.

            EXAMPLES::

                sage: R.<x> = QQ[]
                sage: R._test_quo_rem()

            .. SEEALSO::

                :meth:`_test_euclidean_degree`
            """
            tester = self._tester(**options)
            S = tester.some_elements()
            from sage.misc.misc import some_tuples
            for a,b in some_tuples(S, 2, tester._max_runs):
                if b.is_zero():
                    tester.assertRaises(ZeroDivisionError, lambda: a.quo_rem(b))
                else:
                    q,r = a.quo_rem(b)
                    tester.assertIn(q, self)
                    tester.assertIn(r, self)
                    tester.assertEqual(a,q*b+r)
                    if r != 0:
                        tester.assertLess(r.euclidean_degree(), b.euclidean_degree())

    class ElementMethods:
        @abstract_method
        def euclidean_degree(self):
            r"""
            Return the degree of this element as an element of a Euclidean
            domain, i.e., for elements `a`, `b` the euclidean degree `f`
            satisfies the usual properties:

            1. if `b` is not zero, then there are elements `q` and `r` such
               that `a = bq + r` with `r = 0` or `f(r) < f(b)`
            2. if `a,b` are not zero, then `f(a) \leq f(ab)`

            .. NOTE::

                The name ``euclidean_degree`` was chosen because the euclidean
                function has different names in different contexts, e.g.,
                absolute value for integers, degree for polynomials.

            OUTPUT:

            For nonzero elements, a natural number. For the zero element, this
            might raise an exception or produce some other output, depending on
            the implementation.

            EXAMPLES::

                sage: R.<x> = QQ[]
                sage: x.euclidean_degree()
                1
                sage: ZZ.one().euclidean_degree()
                1
            """

        @coerce_binop
        def gcd(self, other):
            """
            Return the greatest common divisor of this element and ``other``.

            INPUT:

            - ``other`` -- an element in the same ring as ``self``

            ALGORITHM:

            Algorithm 3.2.1 in [Coh1993]_.

            EXAMPLES::

                sage: R.<x> = PolynomialRing(QQ, sparse=True)
                sage: EuclideanDomains().element_class.gcd(x,x+1)
                -1
            """
            A = self
            B = other
            while not B.is_zero():
                Q, R = A.quo_rem(B)
                A = B
                B = R
            return A

        @abstract_method
        def quo_rem(self, other):
            r"""
            Return the quotient and remainder of the division of this element
            by the nonzero element ``other``.

            INPUT:

            - ``other`` -- an element in the same euclidean domain

            OUTPUT: a pair of elements

            EXAMPLES::

                sage: R.<x> = QQ[]
                sage: x.quo_rem(x)
                (1, 0)
            """



# sage_setup: distribution = sagemath-categories
from sage.structure.parent_gens cimport ParentWithGens

cpdef bint _is_Field(x) except -2

cdef class Ring(ParentWithGens):
    cdef public object _zero_element
    cdef public object _one_element


cdef class CommutativeRing(Ring):
    cdef public object __fraction_field

cdef class IntegralDomain(CommutativeRing):
    pass

cdef class DedekindDomain(CommutativeRing):
    pass

cdef class PrincipalIdealDomain(CommutativeRing):
    pass

cdef class Field(CommutativeRing):
    pass

cdef class Algebra(Ring):
    pass

cdef class CommutativeAlgebra(CommutativeRing):
    pass



# sage_setup: distribution = sagemath-categories
r"""
Kac-Moody Algebras With Triangular Decomposition Basis

AUTHORS:

- Travis Scrimshaw (07-15-2017): Initial implementation
"""

# ****************************************************************************
#       Copyright (C) 2017 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.abstract_method import abstract_method
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.categories.category_types import Category_over_base_ring
from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring
from sage.categories.kac_moody_algebras import KacMoodyAlgebras


class TriangularKacMoodyAlgebras(Category_over_base_ring):
    r"""
    Category of Kac-Moody algebras with a distinguished basis that
    respects the triangular decomposition.

    We require that the grading group is the root lattice of the
    appropriate Cartan type.
    """
    @cached_method
    def super_categories(self):
        """
        EXAMPLES::

            sage: from sage.categories.triangular_kac_moody_algebras import TriangularKacMoodyAlgebras
            sage: TriangularKacMoodyAlgebras(QQ).super_categories()
            [Join of Category of graded Lie algebras with basis over Rational Field
                 and Category of kac moody algebras over Rational Field]
        """
        # We do not also derive from (Magmatic) algebras since we don't want *
        #   to be our Lie bracket
        # Also this doesn't inherit the ability to add axioms like Associative
        #   and Unital, both of which do not make sense for Lie algebras
        return [KacMoodyAlgebras(self.base_ring()).WithBasis().Graded()]

    class ParentMethods:
        def _part_on_basis(self, m):
            """
            Return whether the basis element indexed by ``m`` is
            in the lower, zero, or upper part of ``self``.

            OUTPUT:

            `-1` if ``m`` is a negative root, `0` if zero, or `1`
            if ``m`` is a positive root

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: L = lie_algebras.so(QQ, 5)
                sage: L.f()  # indirect doctest
                Finite family {1: E[-alpha[1]], 2: E[-alpha[2]]}
                sage: f1 = L.f(1); f1  # indirect doctest
                E[-alpha[1]]
                sage: L._part_on_basis(f1.leading_support())
                -1
                sage: L._part_on_basis(L.e(1).leading_support())
                1
                sage: all(L._part_on_basis(b) == 0 for b in L.basis() if b.degree() == 0)
                True
            """
            deg = self.degree_on_basis(m)
            if not deg:
                return 0
            return 1 if deg.is_positive_root() else -1

        @cached_method
        def _part_generators(self, positive=False):
            r"""
            Return the Lie algebra generators for the positive or
            negative half of ``self``.

            .. NOTE::

                If the positive/negative generators correspond to the
                generators with (negative) simple roots, then this method
                will find them. If they do not, then this method *must*
                be overwritten. One should also overwrite this method in
                object classes when there is a better method to obtain them.
                Furthermore, this assumes that :meth:`lie_algebra_generators`
                is a finite set.

            INPUT:

            - ``positive`` -- boolean (default: ``False``); if ``True``
              then return positive part generators, otherwise the return
              the negative part generators

            OUTPUT:

            A :func:`~sage.sets.family.Family` whose keys are the
            index set of ``self``.

            EXAMPLES::

                sage: L = LieAlgebra(QQ, cartan_type=['E', 6])                          # needs sage.combinat sage.modules
                sage: list(L._part_generators(False))                                   # needs sage.combinat sage.modules
                [E[-alpha[1]], E[-alpha[2]], E[-alpha[3]],
                 E[-alpha[4]], E[-alpha[5]], E[-alpha[6]]]
            """
            I = self._cartan_type.index_set()
            P = self._cartan_type.root_system().root_lattice()
            ali = P.simple_roots().inverse_family()
            if positive:
                d = {ali[g.degree()]: g for g in self.lie_algebra_generators()
                     if self._part(g) > 0}
            if not positive:
                d = {ali[-g.degree()]: g for g in self.lie_algebra_generators()
                     if self._part(g) < 0}
            from sage.sets.family import Family
            return Family(I, d.__getitem__)

        def e(self, i=None):
            """
            Return the generators `e` of ``self``.

            INPUT:

            - ``i`` -- (optional) if specified, return just the
              generator `e_i`

            EXAMPLES::

                sage: L = lie_algebras.so(QQ, 5)                                        # needs sage.combinat sage.modules
                sage: L.e()                                                             # needs sage.combinat sage.modules
                Finite family {1: E[alpha[1]], 2: E[alpha[2]]}
                sage: L.e(1)                                                            # needs sage.combinat sage.modules
                E[alpha[1]]
            """
            E = self._part_generators(True)
            if i is None:
                return E
            return E[i]

        def f(self, i=None):
            """
            Return the generators `f` of ``self``.

            INPUT:

            - ``i`` -- (optional) if specified, return just the
              generator `f_i`

            EXAMPLES::

                sage: L = lie_algebras.so(QQ, 5)                                        # needs sage.combinat sage.modules
                sage: L.f()                                                             # needs sage.combinat sage.modules
                Finite family {1: E[-alpha[1]], 2: E[-alpha[2]]}
                sage: L.f(1)                                                            # needs sage.combinat sage.modules
                E[-alpha[1]]
            """
            F = self._part_generators(False)
            if i is None:
                return F
            return F[i]

        @abstract_method
        def _negative_half_index_set(self):
            """
            Return an indexing set for the negative half of ``self``.

            EXAMPLES::

                sage: L = lie_algebras.so(QQ, 5)                                        # needs sage.combinat sage.modules
                sage: L._negative_half_index_set()                                      # needs sage.combinat sage.modules
                [-alpha[2], -alpha[1], -alpha[1] - alpha[2],
                 -alpha[1] - 2*alpha[2]]
            """

        @abstract_method
        def _weight_action(self, m, wt):
            """
            Return the action of the basis element indexed by ``m`` on ``wt``.

            INPUT:

            - ``m`` -- an index of a basis element of the Cartan subalgebra
            - ``wt`` -- a weight

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: L = lie_algebras.sp(QQ, 6)
                sage: La = L.cartan_type().root_system().weight_space().fundamental_weights()
                sage: mu = La[1] - 3/5*La[2]
                sage: ac = L.cartan_type().root_system().coroot_lattice().simple_roots()
                sage: L._weight_action(ac[1], mu)
                1
                sage: L._weight_action(ac[2], mu)
                -3/5
                sage: L._weight_action(ac[3], mu)
                0
            """

        def _triangular_key(self, x):
            r"""
            Return a key for sorting for the index ``x`` that respects
            the triangular decomposition by `U^-, U^0, U^+`.

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: L = lie_algebras.sl(QQ, 3)
                sage: La = L.cartan_type().root_system().weight_lattice().fundamental_weights()
                sage: sorted(L.basis().keys(), key=L._basis_key)
                [alpha[2], alpha[1], alpha[1] + alpha[2],
                 alphacheck[1], alphacheck[2],
                 -alpha[2], -alpha[1], -alpha[1] - alpha[2]]
                sage: sorted(L.basis().keys(), key=L._triangular_key)
                [-alpha[2], -alpha[1], -alpha[1] - alpha[2],
                 alphacheck[1], alphacheck[2],
                 alpha[2], alpha[1], alpha[1] + alpha[2]]
            """
            try:
                return (self._part_on_basis(x), self._basis_key(x))
            except AttributeError:
                return (self._part_on_basis(x), x)

        def verma_module(self, la, basis_key=None, **kwds):
            """
            Return the Verma module with highest weight ``la``
            over ``self``.

            INPUT:

            - ``basis_key`` -- (optional) a key function for the indexing
              set of the basis elements of ``self``

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: L = lie_algebras.sl(QQ, 3)
                sage: P = L.cartan_type().root_system().weight_lattice()
                sage: La = P.fundamental_weights()
                sage: M = L.verma_module(La[1] + La[2])
                sage: M
                Verma module with highest weight Lambda[1] + Lambda[2]
                 of Lie algebra of ['A', 2] in the Chevalley basis
            """
            from sage.algebras.lie_algebras.verma_module import VermaModule
            return VermaModule(self, la, basis_key=basis_key, **kwds)

        def simple_module(self, la, basis_key=None, **kwds):
            r"""
            Return the simple module with highest weight ``la``
            over ``self``.

            INPUT:

            - ``la`` -- the weight
            - ``basis_key`` -- (optional) a key function for the indexing
              set of the basis elements of ``self``

            OUTPUT:

            If ``la`` is :meth:`Verma antidominant
            <sage.combinat.root_system.root_lattice_realizations.RootLatticeRealizations.ElementMethods.is_verma_dominant>`,
            then this returns the
            :class:`~sage.algebras.lie_algebras.verma_module.VermaModule`
            of highest weight ``la`` (which is simple and more efficient).
            Otherwise this returns a
            :class:`~sage.algebras.lie_algebras.bgg_dual_module.SimpleModule`.

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: g = lie_algebras.sl(QQ, 3)
                sage: P = g.cartan_type().root_system().weight_lattice()
                sage: La = P.fundamental_weights()
                sage: L = g.simple_module(La[1] + La[2])
                sage: L
                Simple module with highest weight Lambda[1] + Lambda[2]
                 of Lie algebra of ['A', 2] in the Chevalley basis
                sage: L = g.simple_module(-La[1] - La[2])
                sage: L
                Verma module with highest weight -Lambda[1] - Lambda[2]
                 of Lie algebra of ['A', 2] in the Chevalley basis
            """
            if la.is_verma_dominant(positive=False):
                from sage.algebras.lie_algebras.verma_module import VermaModule
                return VermaModule(self, la, basis_key=basis_key, **kwds)

            from sage.algebras.lie_algebras.bgg_dual_module import SimpleModule
            return SimpleModule(self, la, basis_key=basis_key, **kwds)

    class ElementMethods:
        def part(self):
            """
            Return whether the element ``v`` is in the lower,
            zero, or upper part of ``self``.

            OUTPUT:

            `-1` if ``v`` is in the lower part, `0` if in the
            zero part, or `1` if in the upper part

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: L = LieAlgebra(QQ, cartan_type='F4')
                sage: L.inject_variables()
                Defining e1, e2, e3, e4, f1, f2, f3, f4, h1, h2, h3, h4
                sage: e1.part()
                1
                sage: f4.part()
                -1
                sage: (h2 + h3).part()
                0
                sage: (f1.bracket(f2) + 4*f4).part()
                -1
                sage: (e1 + f1).part()
                Traceback (most recent call last):
                ...
                ValueError: element is not in one part
            """
            P = self.parent()
            S = [P._part_on_basis(m) for m in self.support()]
            if all(k < 0 for k in S):
                return -1
            if all(k > 0 for k in S):
                return 1
            if all(k == 0 for k in S):
                return 0
            raise ValueError("element is not in one part")

    class FiniteDimensional(CategoryWithAxiom_over_base_ring):
        r"""
        Category of finite dimensional Kac-Moody algebras (which correspond
        to semisimple Lie algebras) with a distinguished basis that
        respects the triangular decomposition.
        """
        class ParentMethods:
            @lazy_attribute
            def _transpose_basis_mapping(self):
                """
                The mapping on basis elements for the transpose map.

                EXAMPLES::

                    sage: # needs sage.combinat sage.modules
                    sage: g = LieAlgebra(QQ, cartan_type=['A', 2])
                    sage: g._transpose_basis_mapping
                    {-alpha[1]: alpha[1],
                     -alpha[1] - alpha[2]: alpha[1] + alpha[2],
                     -alpha[2]: alpha[2],
                     alpha[1]: -alpha[1],
                     alpha[1] + alpha[2]: -alpha[1] - alpha[2],
                     alpha[2]: -alpha[2],
                     alphacheck[1]: alphacheck[1],
                     alphacheck[2]: alphacheck[2]}
                """
                K = self.basis().keys()
                deg_map = {}
                ret = {}
                for k in K:
                    deg = self.degree_on_basis(k)
                    if deg:
                        deg_map[deg] = k
                    else:
                        ret[k] = k
                for al, k in deg_map.items():
                    ret[k] = deg_map[-al]
                return ret

            @cached_method
            def _transpose_on_basis(self, m):
                r"""
                Return the transpose map on the basis element indexed by ``m``.

                EXAMPLES::

                    sage: # needs sage.combinat sage.modules
                    sage: g = LieAlgebra(QQ, cartan_type=['B', 2])
                    sage: B = g.basis()
                    sage: [(B[k], g._transpose_on_basis(k)) for k in B.keys()]
                    [(E[alpha[2]], E[-alpha[2]]),
                     (E[alpha[1]], E[-alpha[1]]),
                     (E[alpha[1] + alpha[2]], E[-alpha[1] - alpha[2]]),
                     (E[alpha[1] + 2*alpha[2]], E[-alpha[1] - 2*alpha[2]]),
                     (h1, h1),
                     (h2, h2),
                     (E[-alpha[2]], E[alpha[2]]),
                     (E[-alpha[1]], E[alpha[1]]),
                     (E[-alpha[1] - alpha[2]], E[alpha[1] + alpha[2]]),
                     (E[-alpha[1] - 2*alpha[2]], E[alpha[1] + 2*alpha[2]])]
                """
                return self.monomial(self._transpose_basis_mapping[m])

            @lazy_attribute
            def transpose(self):
                r"""
                The transpose map of ``self``.

                EXAMPLES::

                    sage: # needs sage.combinat sage.modules
                    sage: g = LieAlgebra(QQ, cartan_type=['B', 2])
                    sage: g.transpose
                    Generic endomorphism of Lie algebra of ['B', 2] in the Chevalley basis
                """
                return self.module_morphism(self._transpose_on_basis, codomain=self)

        class ElementMethods:
            def transpose(self):
                r"""
                Return the transpose of ``self``.

                The transpose `\tau` is the map that sends the root basis
                elements `e_{\alpha} \leftrightarrow e_{-\alpha}` and fixes
                the Cartan subalgebra `h_{\alpha}`. It is an anti-involution
                in the sense `[\tau(a), \tau(b)] = \tau([b, a])`.

                EXAMPLES::

                    sage: # needs sage.combinat sage.modules
                    sage: g = LieAlgebra(QQ, cartan_type=['G', 2])
                    sage: for b in g.basis():
                    ....:     for bp in g.basis():
                    ....:         a = g[b.transpose(), bp.transpose()]
                    ....:         ap = g[bp, b].transpose()
                    ....:         assert a == ap
                """
                return self.parent().transpose(self)



# sage_setup: distribution = sagemath-categories
r"""
Lie Algebras With Basis

AUTHORS:

- Travis Scrimshaw (07-15-2013): Initial implementation
"""

#*****************************************************************************
#       Copyright (C) 2013-2024 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from sage.misc.abstract_method import abstract_method
from sage.misc.lazy_import import LazyImport
from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring
from sage.categories.lie_algebras import LieAlgebras


class LieAlgebrasWithBasis(CategoryWithAxiom_over_base_ring):
    """
    Category of Lie algebras with a basis.
    """
    _base_category_class_and_axiom = (LieAlgebras, "WithBasis")

    def example(self, gens=None):
        """
        Return an example of a Lie algebra as per
        :meth:`Category.example <sage.categories.category.Category.example>`.

        EXAMPLES::

            sage: LieAlgebras(QQ).WithBasis().example()                                 # needs sage.combinat sage.modules
            An example of a Lie algebra: the abelian Lie algebra on the
             generators indexed by Partitions over Rational Field

        Another set of generators can be specified as an optional argument::

            sage: LieAlgebras(QQ).WithBasis().example(Compositions())                   # needs sage.combinat sage.modules
            An example of a Lie algebra: the abelian Lie algebra on the
             generators indexed by Compositions of nonnegative integers
             over Rational Field
        """
        if gens is None:
            from sage.combinat.partition import Partitions
            gens = Partitions()
        from sage.categories.examples.lie_algebras_with_basis import Example
        return Example(self.base_ring(), gens)

    Graded = LazyImport('sage.categories.graded_lie_algebras_with_basis',
                        'GradedLieAlgebrasWithBasis',
                        as_name='Graded')

    class ParentMethods:
        def _basis_key(self, x):
            """
            Return the key used to compare two basis element indices.

            The default is to call the element itself.

            TESTS::

                sage: L = LieAlgebras(QQ).WithBasis().example()                         # needs sage.combinat sage.modules
                sage: L._basis_key(Partition([3,1]))                                    # needs sage.combinat sage.modules
                [3, 1]
            """
            return x

        @abstract_method(optional=True)
        def bracket_on_basis(self, x, y):
            """
            Return the bracket of basis elements indexed by ``x`` and ``y``
            where ``x < y``. If this is not implemented, then the method
            ``_bracket_()`` for the elements must be overwritten.

            EXAMPLES::

                sage: L = LieAlgebras(QQ).WithBasis().example()                         # needs sage.combinat sage.modules
                sage: L.bracket_on_basis(Partition([3,1]), Partition([2,2,1,1]))        # needs sage.combinat sage.modules
                0
            """

        def module(self):
            """
            Return an `R`-module which is isomorphic to the
            underlying `R`-module of ``self``.

            See
            :meth:`sage.categories.lie_algebras.LieAlgebras.module` for
            an explanation.

            EXAMPLES::

                sage: L = LieAlgebras(QQ).WithBasis().example()                         # needs sage.combinat sage.modules
                sage: L.module()                                                        # needs sage.combinat sage.modules
                Free module generated by Partitions over Rational Field
            """
            from sage.combinat.free_module import CombinatorialFreeModule
            try:
                # Try to see if it has an indexing set
                return CombinatorialFreeModule(self.base_ring(), self.basis().keys())
            except AttributeError:
                # Otherwise just index by the basis of ``self`` as a fallback
                return CombinatorialFreeModule(self.base_ring(), self.basis())

        def from_vector(self, v, order=None, coerce=False):
            """
            Return the element of ``self`` corresponding to the
            vector ``v`` in ``self.module()``.

            Implement this if you implement :meth:`module`; see the
            documentation of
            :meth:`sage.categories.lie_algebras.LieAlgebras.module`
            for how this is to be done.

            EXAMPLES::

                sage: L = LieAlgebras(QQ).FiniteDimensional().WithBasis().example()     # needs sage.modules
                sage: u = L.from_vector(vector(QQ, (1, 0, 0))); u                       # needs sage.modules
                (1, 0, 0)
                sage: parent(u) is L                                                    # needs sage.modules
                True
            """
            B = self.basis()
            return self.sum(v[i] * B[i] for i in v.support())

        # Remove once #22629 is merged
        def dimension(self):
            """
            Return the dimension of ``self``.

            EXAMPLES::

                sage: L = LieAlgebras(QQ).FiniteDimensional().WithBasis().example()     # needs sage.modules
                sage: L.dimension()                                                     # needs sage.modules
                3

            ::

                sage: L = LieAlgebra(QQ, 'x,y', {('x','y'): {'x':1}})                   # needs sage.combinat sage.modules
                sage: L.dimension()                                                     # needs sage.combinat sage.modules
                2
            """
            return self.basis().cardinality()

        def pbw_basis(self, basis_key=None, **kwds):
            """
            Return the Poincare-Birkhoff-Witt basis of the universal
            enveloping algebra corresponding to ``self``.

            EXAMPLES::

                sage: L = lie_algebras.sl(QQ, 2)                                        # needs sage.combinat sage.modules
                sage: PBW = L.pbw_basis()                                               # needs sage.combinat sage.modules
            """
            from sage.algebras.lie_algebras.poincare_birkhoff_witt \
                import PoincareBirkhoffWittBasis
            return PoincareBirkhoffWittBasis(self, basis_key, **kwds)

        poincare_birkhoff_witt_basis = pbw_basis

        _construct_UEA = pbw_basis

    class ElementMethods:
        def _bracket_(self, y):
            """
            Return the Lie bracket ``[self, y]``, where ``y`` is an
            element of the same Lie algebra as ``self``.

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: L = LieAlgebras(QQ).WithBasis().example()
                sage: G = L.lie_algebra_generators()
                sage: x = G[Partition([4,3,3,1])]
                sage: y = G[Partition([6,1])]
                sage: x.bracket(y)
                0
            """
            P = self.parent()

            def term(ml, mr):
                key_ml = P._basis_key(ml)
                key_mr = P._basis_key(mr)
                if key_ml == key_mr:
                    return P.zero()
                if key_ml < key_mr:
                    return P.bracket_on_basis(ml, mr)
                return -P.bracket_on_basis(mr, ml)

            return P.sum(cl * cr * term(ml, mr)
                         for ml, cl in self for mr, cr in y)

        def to_vector(self, order=None):
            """
            Return the vector in ``g.module()`` corresponding to the
            element ``self`` of ``g`` (where ``g`` is the parent of
            ``self``).

            Implement this if you implement ``g.module()``.
            See :meth:`sage.categories.lie_algebras.LieAlgebras.module`
            for how this is to be done.

            EXAMPLES::

                sage: L = LieAlgebras(QQ).FiniteDimensional().WithBasis().example()     # needs sage.modules
                sage: L.an_element().to_vector()                                        # needs sage.modules
                (0, 0, 0)

            .. TODO::

                Doctest this implementation on an example not overshadowed.
            """
            M = self.parent().module()
            B = M.basis()
            return M.sum(self[i] * B[i] for i in self.support())

        def lift(self):
            """
            Lift ``self`` to the universal enveloping algebra.

            EXAMPLES::

                sage: # needs sage.combinat sage.groups
                sage: S = SymmetricGroup(3).algebra(QQ)
                sage: L = LieAlgebra(associative=S)
                sage: x = L.gen(3)
                sage: y = L.gen(1)
                sage: x.lift()
                b3
                sage: y.lift()
                b1
                sage: x * y
                b1*b3 + b4 - b5
            """
            P = self.parent()
            UEA = P.universal_enveloping_algebra()
            s = UEA.zero()
            if not self:
                return s
            # Special hook for when the index set of the parent of ``self``
            #   does not match the generators index set of the UEA.
            if hasattr(P, '_UEA_names_map'):
                names_map = P._UEA_names_map
                gen_dict = UEA.gens_dict()
                for t, c in self.monomial_coefficients(copy=False).items():
                    s += c * gen_dict[names_map[t]]
            else:
                try:
                    gen_dict = UEA.algebra_generators()
                except (TypeError, AttributeError):
                    gen_dict = UEA.gens_dict()
                for t, c in self.monomial_coefficients(copy=False).items():
                    s += c * gen_dict[t]
            return s



# sage_setup: distribution = sagemath-categories
r"""
Lattice posets
"""
#*****************************************************************************
#  Copyright (C) 2011 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.misc.cachefunc import cached_method
from sage.misc.abstract_method import abstract_method
from sage.misc.lazy_import import LazyImport
from sage.categories.category import Category
from sage.categories.posets import Posets


class LatticePosets(Category):
    r"""
    The category of lattices, i.e. partially ordered sets in which any
    two elements have a unique supremum (the elements' least upper
    bound; called their *join*) and a unique infimum (greatest lower bound;
    called their *meet*).

    EXAMPLES::

        sage: LatticePosets()
        Category of lattice posets
        sage: LatticePosets().super_categories()
        [Category of posets]
        sage: LatticePosets().example()
        NotImplemented

    .. SEEALSO:: :class:`~sage.categories.posets.Posets`, :class:`FiniteLatticePosets`, :func:`LatticePoset`

    TESTS::

        sage: C = LatticePosets()
        sage: TestSuite(C).run()
    """
    @cached_method
    def super_categories(self):
        r"""
        Return a list of the (immediate) super categories of
        ``self``, as per :meth:`Category.super_categories`.

        EXAMPLES::

            sage: LatticePosets().super_categories()
            [Category of posets]
        """
        return [Posets()]

    Finite = LazyImport('sage.categories.finite_lattice_posets', 'FiniteLatticePosets')

    class ParentMethods:

        @abstract_method
        def meet(self, x, y):
            """
            Return the meet of `x` and `y` in this lattice.

            INPUT:

            - ``x``, ``y`` -- elements of ``self``

            EXAMPLES::

                sage: D = LatticePoset((divisors(30), attrcall("divides")))             # needs sage.graphs sage.modules
                sage: D.meet( D(6), D(15) )                                             # needs sage.graphs sage.modules
                3
            """

        @abstract_method
        def join(self, x, y):
            """
            Return the join of `x` and `y` in this lattice.

            INPUT:

            - ``x``, ``y`` -- elements of ``self``

            EXAMPLES::

                sage: D = LatticePoset((divisors(60), attrcall("divides")))             # needs sage.graphs sage.modules
                sage: D.join( D(6), D(10) )                                             # needs sage.graphs sage.modules
                30
            """



# sage_setup: distribution = sagemath-categories
r"""
Graded Hopf algebras
"""
# ****************************************************************************
#  Copyright (C) 2008      Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#                2008-2011 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************


def GradedHopfAlgebras(base_ring):
    r"""
    The category of graded Hopf algebras.

    EXAMPLES::

        sage: C = GradedHopfAlgebras(QQ); C
        Join of Category of Hopf algebras over Rational Field
            and Category of graded algebras over Rational Field
            and Category of graded coalgebras over Rational Field
        sage: C is HopfAlgebras(QQ).Graded()
        True

    TESTS::

        sage: TestSuite(C).run()

    .. NOTE::

        This is not a graded Hopf algebra as is typically defined
        in algebraic topology as the product in the tensor square
        `(x \otimes y) (a \otimes b) = (xa) \otimes (yb)` does
        not carry an additional sign. For this, instead use
        :class:`super Hopf algebras
        <sage.categories.hopf_algebras.HopfAlgebras.Super>`.
    """
    from sage.categories.hopf_algebras import HopfAlgebras
    return HopfAlgebras(base_ring).Graded()



# sage_setup: distribution = sagemath-categories
r"""
R-trivial semigroups
"""
#*****************************************************************************
#  Copyright (C) 2016 Nicolas M. Thiéry <nthiery at users.sf.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.categories.category_with_axiom import CategoryWithAxiom
from sage.categories.semigroups import Semigroups


class RTrivialSemigroups(CategoryWithAxiom):
    def extra_super_categories(self):
        r"""
        Implement the fact that a `R`-trivial semigroup is `H`-trivial.

        EXAMPLES::

            sage: Semigroups().RTrivial().extra_super_categories()
            [Category of h trivial semigroups]
        """
        return [Semigroups().HTrivial()]

    def Commutative_extra_super_categories(self):
        r"""
        Implement the fact that a commutative `R`-trivial semigroup is `J`-trivial.

        EXAMPLES::

            sage: Semigroups().RTrivial().Commutative_extra_super_categories()
            [Category of j trivial semigroups]

        TESTS::

            sage: Semigroups().RTrivial().Commutative() is Semigroups().JTrivial().Commutative()
            True
        """
        return [self.JTrivial()]



r"""
Kähler Algebras

AUTHORS:

- Shriya M
"""
# ****************************************************************************
#       Copyright (C) 2024 Shriya M <25shriya at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.category_types import Category_over_base_ring
from sage.categories.graded_algebras_with_basis import GradedAlgebrasWithBasis
from sage.misc.abstract_method import abstract_method
from sage.quadratic_forms.quadratic_form import QuadraticForm


class KahlerAlgebras(Category_over_base_ring):
    r"""
    The category of graded algebras satisfying the Kähler package.

    A finite-dimensional graded algebra `\bigoplus_{k=1}^{r}A^k` satisfies
    the *Kähler package* if the following properties hold:

    - Poincaré duality: There exists a perfect `\ZZ`-bilinear pairing
      given by

      .. MATH::

          A^k \times A^{r-k} \longrightarrow \ZZ \\
            (a,b) \mapsto \deg(a \cdot b).

    - Hard-Lefschetz Theorem: The graded algebra contains *Lefschetz elements*
      `\omega \in A^{1}_{\RR}` such that multiplication by `\omega` is
      an injection from `A^k_{\RR} \longrightarrow A^{k+1}_{\RR}`
      for all `k < \frac{r}{2}`.

    - Hodge-Riemann-Minikowski Relations: Every Lefchetz element `\omega`,
      define quadratic forms on `A^{k}_{\RR}` given by

      .. MATH::

          a \mapsto (-1)^k \deg(a \cdot \omega^{r-2k} \cdot a)

      This quadratic form becomes positive definite upon restriction to the
      kernel of the following map

      .. MATH::

          A^k_\RR \longrightarrow A^{r-k+1}_\RR \\
                a \mapsto a \cdot \omega^{r-2k+1}.

    REFERENCES:

    - [ANR2023]_

    TESTS::

        sage: from sage.categories.kahler_algebras import KahlerAlgebras

        sage: C = KahlerAlgebras(QQ)
        sage: TestSuite(C).run()
    """
    def super_categories(self):
        r"""
        Return the super categories of ``self``.

        EXAMPLES::

            sage: from sage.categories.kahler_algebras import KahlerAlgebras

            sage: C = KahlerAlgebras(QQ); C
            Category of kahler algebras over Rational Field
            sage: sorted(C.super_categories(), key=str)
            [Category of finite dimensional graded algebras with basis over
            Rational Field]
        """
        return [GradedAlgebrasWithBasis(self.base_ring()).FiniteDimensional()]

    class ParentMethods:
        @abstract_method
        def poincare_pairing(self, a, b):
            r"""
            Return the Poincaré pairing of two elements of the Kähler algebra.

            EXAMPLES::

                sage: ch = matroids.catalog.Fano().chow_ring(QQ, True, 'fy')
                sage: Ba, Bb, Bc, Bd, Be, Bf, Bg, Babf, Bace, Badg, Bbcd, Bbeg, Bcfg, Bdef, Babcdefg = ch.gens()[8:]
                sage: u = ch(-Babf^2 + Bcfg^2 - 8/7*Bc*Babcdefg + 1/2*Bd*Babcdefg - Bf*Babcdefg - Bg*Babcdefg); u
                -Babf^2 + Bcfg^2 - 8/7*Bc*Babcdefg + 1/2*Bd*Babcdefg - Bf*Babcdefg - Bg*Babcdefg
                sage: v = ch(Bg - 2/37*Babf + Badg + Bbeg + Bcfg + Babcdefg); v
                Bg - 2/37*Babf + Badg + Bbeg + Bcfg + Babcdefg
                sage: ch.poincare_pairing(v, u)
                3
            """

        @abstract_method
        def lefschetz_element(self):
            r"""
            Return one Lefschetz element of the given Kähler algebra.

            EXAMPLES::

                sage: U46 = matroids.Uniform(4, 6)
                sage: C = U46.chow_ring(QQ, False)
                sage: w = C.lefschetz_element(); w
                -2*A01 - 2*A02 - 2*A03 - 2*A04 - 2*A05 - 2*A12 - 2*A13 - 2*A14
                - 2*A15 - 2*A23 - 2*A24 - 2*A25 - 2*A34 - 2*A35 - 2*A45 - 6*A012
                - 6*A013 - 6*A014 - 6*A015 - 6*A023 - 6*A024 - 6*A025 - 6*A034
                - 6*A035 - 6*A045 - 6*A123 - 6*A124 - 6*A125 - 6*A134 - 6*A135
                - 6*A145 - 6*A234 - 6*A235 - 6*A245 - 6*A345 - 30*A012345
                sage: basis_deg = {}
                sage: for b in C.basis():
                ....:     deg = b.homogeneous_degree()
                ....:     if deg not in basis_deg:
                ....:         basis_deg[deg] = []
                ....:     basis_deg[deg].append(b)
                sage: m = max(basis_deg); m
                3
                sage: len(basis_deg[1]) == len(basis_deg[2])
                True
                sage: matrix([(w*b).to_vector() for b in basis_deg[1]]).rank()
                36
                sage: len(basis_deg[2])
                36
            """

        def hodge_riemann_relations(self, k):
            r"""
            Return the quadratic form for the corresponding ``k``
            (`< \frac{r}{2}`) for the Kähler algebra, where `r` is the top degree.

            EXAMPLES::

                sage: ch = matroids.Uniform(4, 6).chow_ring(QQ, False)
                sage: ch.hodge_riemann_relations(1)
                Quadratic form in 36 variables over Rational Field with coefficients:
                [ 3 -1 -1 3 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 3 ]
                [ * 3 -1 3 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 3 -1 3 ]
                [ * * 3 3 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 3 3 ]
                [ * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * 3 -1 3 -1 3 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 3 ]
                [ * * * * * 3 3 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 3 -1 -1 3 ]
                [ * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * 3 3 3 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 3 ]
                [ * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * 3 -1 3 -1 -1 3 -1 -1 3 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 3 ]
                [ * * * * * * * * * * * 3 3 -1 3 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 3 ]
                [ * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * 3 3 3 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 3 -1 -1 -1 -1 -1 -1 -1 3 ]
                [ * * * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * 3 3 3 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 3 -1 -1 -1 3 ]
                [ * * * * * * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * * * 3 -1 3 -1 3 -1 -1 -1 -1 3 -1 -1 -1 3 -1 3 ]
                [ * * * * * * * * * * * * * * * * * * * * * 3 3 -1 -1 3 -1 -1 3 -1 -1 -1 3 -1 -1 3 ]
                [ * * * * * * * * * * * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * 3 3 3 -1 3 -1 -1 -1 3 -1 -1 -1 3 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * * * 3 3 3 3 -1 -1 -1 -1 3 3 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 3 -1 -1 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 3 3 3 3 3 3 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 3 -1 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 3 -1 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 3 -1 -1 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 3 -1 ]
                [ * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 3 ]
                sage: ch.hodge_riemann_relations(3)
                Traceback (most recent call last):
                ...
                ValueError: k must be less than r/2 < 2
            """
            r = self.top_degree()
            if k > (r/2):
                raise ValueError("k must be less than r/2 < 2")
            basis_k = []
            lefschetz_el = self.lefschetz_element()
            for b in self.basis():
                if b.homogeneous_degree() == k:
                    basis_k.append(b)
            coeff = []
            for i,el in enumerate(basis_k):
                for j in range(i, len(basis_k)):
                    coeff.append((el * (lefschetz_el ** (r-(2*k)) * basis_k[j])).degree())
            return QuadraticForm(self.base_ring(), len(basis_k), coeff)



# sage_setup: distribution = sagemath-categories
r"""
Dedekind Domains
"""
# ****************************************************************************
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************
from sage.categories.category import Category
from sage.categories.integral_domains import IntegralDomains


class DedekindDomains(Category):
    """
    The category of Dedekind domains.

    A Dedekind domain is a Noetherian integral domain of Krull
    dimension one that is integrally closed in its field of fractions.

    EXAMPLES::

        sage: C = DedekindDomains(); C
        Category of Dedekind domains
        sage: C.super_categories()
        [Category of integral domains]

    TESTS::

        sage: TestSuite(C).run()
    """
    def super_categories(self):
        """
        EXAMPLES::

            sage: DedekindDomains().super_categories()
            [Category of integral domains]
        """
        return [IntegralDomains()]

    class ParentMethods:
        def krull_dimension(self):
            """
            Return 1 since Dedekind domains have Krull dimension 1.

            EXAMPLES:

            The following are examples of Dedekind domains::

                sage: ZZ.krull_dimension()
                1
                sage: x = polygen(ZZ, 'x')
                sage: K = NumberField(x^2 + 1, 's')                                         # needs sage.rings.number_field
                sage: OK = K.ring_of_integers()                                             # needs sage.rings.number_field
                sage: OK.krull_dimension()                                                  # needs sage.rings.number_field
                1
            """
            from sage.rings.integer_ring import ZZ
            return ZZ.one()

        def is_integrally_closed(self) -> bool:
            """
            Return ``True`` since Dedekind domains are integrally closed.

            EXAMPLES:

            The following are examples of Dedekind domains::

                sage: ZZ.is_integrally_closed()
                True
                sage: x = polygen(ZZ, 'x')
                sage: K = NumberField(x^2 + 1, 's')                                         # needs sage.rings.number_field
                sage: OK = K.ring_of_integers()                                             # needs sage.rings.number_field
                sage: OK.is_integrally_closed()                                             # needs sage.rings.number_field
                True

            These, however, are not Dedekind domains::

                sage: QQ.is_integrally_closed()
                True
                sage: S = ZZ[sqrt(5)]; S.is_integrally_closed()                             # needs sage.rings.number_field sage.symbolic
                False
                sage: T.<x,y> = PolynomialRing(QQ, 2); T
                Multivariate Polynomial Ring in x, y over Rational Field
                sage: T.is_integral_domain()
                True
            """
            return True

        def integral_closure(self):
            r"""
            Return ``self`` since Dedekind domains are integrally closed.

            EXAMPLES::

                sage: # needs sage.rings.number_field
                sage: x = polygen(ZZ, 'x')
                sage: K = NumberField(x^2 + 1, 's')
                sage: OK = K.ring_of_integers()
                sage: OK.integral_closure()
                Gaussian Integers generated by s in Number Field in s
                 with defining polynomial x^2 + 1
                sage: OK.integral_closure() == OK
                True

                sage: QQ.integral_closure() == QQ
                True
            """
            return self

        def is_noetherian(self) -> bool:
            r"""
            Return ``True`` since Dedekind domains are Noetherian.

            EXAMPLES:

            The integers, `\ZZ`, and rings of integers of number
            fields are Dedekind domains::

                sage: ZZ.is_noetherian()
                True
                sage: x = polygen(ZZ, 'x')
                sage: K = NumberField(x^2 + 1, 's')                                         # needs sage.rings.number_field
                sage: OK = K.ring_of_integers()                                             # needs sage.rings.number_field
                sage: OK.is_noetherian()                                                    # needs sage.rings.number_field
                True
                sage: QQ.is_noetherian()
                True
            """
            return True

    class ElementMethods:
        pass
